/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
This file implements the SyncGenie's default "all-singing, all-dancing" firmware.
It is complex and very flexible, and as such not necessarily a good model for
learning to program the SyncGenie library. It is probably better to work from the
other example sketches instead.
*/

#include "SyncGenieBoardRevision.h" // explicitly including this here makes it possible to do automated overshadowing of the board revision, e.g. in the `go` script
#include "SyncGenie.h"
SyncGenie genie;

#include "Triggering.h"

typedef enum { KEEP_MODE = -1, DISABLED = 0, SWITCH, AC_TRIGGER, DC_TRIGGER } JackMode;
typedef enum { KEEP_CIRCUIT_TYPE = -1, READ_BOTH = 0, RING_HI_READ_TIP, RING_LO_READ_TIP, TIP_HI_READ_RING, TIP_LO_READ_RING } CircuitType;
typedef enum { NO_JACK    = -2, ALL_JACKS    = -1, TRS1 = 0, TRS2, TRS3, TRS4 } JackIndex;
typedef enum { NO_CHANNEL = -2, ALL_CHANNELS = -1,   T1 = 0, R1, T2, R2, T3, R3, T4, R4 } ChannelIndex;

bool gMeasureIdleLoops = false;
unsigned long gLoopsPerBar = 1;

const int NUMBER_OF_JACKS = 4;
const int NUMBER_OF_CHANNELS = NUMBER_OF_JACKS * 2;

void  *                 gTriggers[ NUMBER_OF_CHANNELS ] = {           NULL,             NULL,           NULL,             NULL,           NULL,             NULL,           NULL,             NULL };

int                    gInputPins[ NUMBER_OF_CHANNELS ] = { genie.TRS1_TIP,  genie.TRS1_RING, genie.TRS2_TIP,  genie.TRS2_RING, genie.TRS3_TIP,  genie.TRS3_RING, genie.TRS4_TIP,  genie.TRS4_RING };
int                  gOutput1Pins[ NUMBER_OF_CHANNELS ] = {   genie.D0_OUT,     genie.D1_OUT,   genie.D2_OUT,     genie.D3_OUT,   genie.D4_OUT,     genie.D5_OUT,   genie.D6_OUT,     genie.D7_OUT };
int                  gOutput2Pins[ NUMBER_OF_CHANNELS ] = {  NO_OUTPUT_PIN,    NO_OUTPUT_PIN,  NO_OUTPUT_PIN,    NO_OUTPUT_PIN,  NO_OUTPUT_PIN,    NO_OUTPUT_PIN,  NO_OUTPUT_PIN,    NO_OUTPUT_PIN };
int                   gThresholds[ NUMBER_OF_CHANNELS ] = {             15,               15,             15,               15,             15,               15,             15,               15 };
unsigned long  gDurationMinMicros[ NUMBER_OF_CHANNELS ] = {          10000,            10000,          10000,            10000,          10000,            10000,          10000,            10000 };
unsigned long  gDurationIncMicros[ NUMBER_OF_CHANNELS ] = {          10000,            10000,          10000,            10000,          10000,            10000,          10000,            10000 };
unsigned long   gRefractoryMillis[ NUMBER_OF_CHANNELS ] = {            200,              200,            200,              200,            200,              200,            200,              200 };
unsigned long gSamplePeriodMicros[ NUMBER_OF_CHANNELS ] = {            160,              160,            160,              160,            160,              160,            160,              160 }; /* see @@@ below */
JackMode            gChannelModes[ NUMBER_OF_CHANNELS ] = {       DISABLED,         DISABLED,       DISABLED,         DISABLED,       DISABLED,         DISABLED,       DISABLED,         DISABLED };
JackMode               gJackModes[ NUMBER_OF_JACKS    ] = {                DISABLED,                         DISABLED,                         DISABLED,                         DISABLED          };
CircuitType         gCircuitTypes[ NUMBER_OF_JACKS    ] = {               READ_BOTH,                        READ_BOTH,                        READ_BOTH,                        READ_BOTH          };
bool               gVisibleInMenu[ NUMBER_OF_JACKS    ] = {                   false,                            false,                            false,                            false          };

/*
	@@@ Default sampling rate: @@@

	On the old rev.2 box, with a Teensy 3.2, the "idle loop" test shows results that vary with some granularity depending on how many channels are active and what the sampling rate is.
    You can light all the bars, none of the bars, and usually two or three steps in between.

	On the rev.4 box with a Teensy 4.1, processing takes 16-17 microseconds per channel per sample. If you have all 8 channels running as AC triggers, 160 micros (6250 Hz) is the fastest you
	can go without the bar graph going empty. If you disable all but one trigger (TRS1 Mode = "AC TRIGGER",  TRS1 Circuit = "RING LO READ TIP",  all other TRS modes = "NO PROCESSING") then
	you can go to 50 kHz (20 micros).
*/

#include "Introspection.h"

#define MODIFIED_LEFT_ALWAYS_GOES_BACK 1

const char MAIN_DOC[] =
	"Turn the selector wheel left or right to navigate menu options, and click the selector to select a menu item. "
	"Menu items in ALL-\tCAPS generally perform an action or re-\tconfiguration immediately when clicked. "
#	if MODIFIED_LEFT_ALWAYS_GOES_BACK
	"You can also press the selector and hold it in while turning: press-\tand-\tturn-\tleft is a shortcut for selecting \"GO BACK\" to navigate back up the menu tree; press-\tand-\tturn-\tright is the same as clicking, "
	"with one exception: if you are adjusting a numeric value like a threshold, and the adjustment delta value is shown on the right, then press-\tand-\tturn-\tright allows you to change the delta. "
#	else
	"If you are adjusting a numeric value such as a threshold, you can also press the selector button and hold it in while turning: this lets you increase or decrease the adjustment delta. "
	"Anywhere else in the menu, holding the selector pressed-\tin while turning left is a shortcut for selecting \"GO BACK\" to navigate back up the menu tree. "
#	endif
	"\n!! SAFETY REMINDER: you must ensure that optical isolation is implemented between the SyncGenie and any sensor circuitry that contacts a human research subject. Most amps/\theadsets do incorporate it, but you must verify this, and build your own isolation solution if not. "
	"\nFull manual: osf.io/dszfa "
	"\nTo exit this doc screen, you can use the press-\tand-\tturn-\tleft gesture, or just click the selector once. "
;
const char IDLE_LOOP_DOC[] =
	"When the idle-\tloop count is turned on, the LED bar graph and D_OUT pins no longer carry trigger pulses. "
	"Instead, the LED bar graph is used as a bar graph to report the number of loop() calls that return without a sample yet being due. "
	"If none of the bars light up (even with Loops Per Bar set to 1), then there are no idle loops and this probably means that the processor cannot keep up with the desired sampling rate. "
	"In this case, you should decrease the number of active channels or decrease the sampling rate. "
;

#include "SerialTriggerOutput.h"
SerialTriggerOutput gSerialTriggerOutput( Serial1, 0 );
bool gUseSerialTriggerOutput = false;
const char SERIAL_TRIGGER_DOC[] = 
	"When Serial Output is activated, the LED bar graph and D_OUT pins no longer carry individual trigger pulses. "
	"Instead, the pin corresponding to D6_OUT (pin 8 of the 25-pin D-\tSUB output) is used to transmit data using a serial protocol at baud 115200. "
	"This is the form in which the FlexEEG headset accepts trigger signals. "
	"The D6 LED will appear to be on all the time (in fact, it is flickering very fast). "
	"\nClick to exit."
;

unsigned long Run( void )
{
	unsigned long now = micros();
	static unsigned long idleLoops = 0;
	bool usedTrigger = false, newSample = false;
	for( int channelIndex = 0; channelIndex < NUMBER_OF_CHANNELS; channelIndex++ )
	{
		bool outputIsHigh = false;
		
		if( gTriggers[ channelIndex ] )
		{
			switch( gChannelModes[ channelIndex ] )
			{
				case AC_TRIGGER:
				case DC_TRIGGER:
				{
					Trigger * t = ( Trigger * )gTriggers[ channelIndex ];
					if( t->read( now ) ) newSample = true;
					outputIsHigh = t->write( now ); // NB: this is depth-first - comment this out and uncomment the section below to do breadth-first (but it doesn't seem to be better)
					usedTrigger = true;
					break;
				}
				case SWITCH:
				{
					Switch * sw = ( Switch * )gTriggers[ channelIndex ];
					int event = sw->event();
					outputIsHigh = ( sw->lastEvent > 0 );
					if( event )
					{
						// NB: this is all being done depth-first, so verify what the differences are (both on the input side and the output side) between the first (TRS1 tip) and last (TRS4 ring)
						if( gOutput1Pins[ channelIndex ] >= 0 && !gMeasureIdleLoops && !gUseSerialTriggerOutput ) digitalWrite( gOutput1Pins[ channelIndex ], outputIsHigh );
						if( gOutput2Pins[ channelIndex ] >= 0 ) digitalWrite( gOutput2Pins[ channelIndex ], outputIsHigh );
					}
					break;
				}
				default:
					break;
			}
		}
		if( gUseSerialTriggerOutput )
		{
			switch( gOutput1Pins[ channelIndex ] )
			{
				case SyncGenie::D0_OUT: gSerialTriggerOutput.setBit( 0, outputIsHigh ); break;
				case SyncGenie::D1_OUT: gSerialTriggerOutput.setBit( 1, outputIsHigh ); break;
				case SyncGenie::D2_OUT: gSerialTriggerOutput.setBit( 2, outputIsHigh ); break;
				case SyncGenie::D3_OUT: gSerialTriggerOutput.setBit( 3, outputIsHigh ); break;
				case SyncGenie::D4_OUT: gSerialTriggerOutput.setBit( 4, outputIsHigh ); break;
				case SyncGenie::D5_OUT: gSerialTriggerOutput.setBit( 5, outputIsHigh ); break;
				case SyncGenie::D6_OUT: gSerialTriggerOutput.setBit( 6, outputIsHigh ); break;
				case SyncGenie::D7_OUT: gSerialTriggerOutput.setBit( 7, outputIsHigh ); break;
				default: break;
			}
		}
	}
	/* // uncomment this to do a more breadth-first sweep of the channels (and comment out the write() call above)
	for( int i = 0; i < NUMBER_OF_CHANNELS; i++ )
	{
		if( !gTriggers[ i ] ) continue;
		if( gChannelModes[ i ] != AC_TRIGGER && gChannelModes[ i ] != DC_TRIGGER ) continue;
		( ( Trigger * )gTriggers[ i ] )->write( now );
	}*/
	if( gUseSerialTriggerOutput )
	{
		gSerialTriggerOutput.send();
	}
	else if( gMeasureIdleLoops )
	{
		if( newSample || !usedTrigger ) idleLoops = 0 * genie.bar( idleLoops / gLoopsPerBar );
		else idleLoops++;
	}
	return now;
}

void SetTriggerParameters( void )
{
	for( int channelIndex = 0; channelIndex < NUMBER_OF_CHANNELS; channelIndex++ )
	{
		if( !gTriggers[ channelIndex ] ) continue;
		if( gChannelModes[ channelIndex ] == AC_TRIGGER || gChannelModes[ channelIndex ] == DC_TRIGGER )
		{
			Trigger * t = ( Trigger * )gTriggers[ channelIndex ];
			t->samplePeriodInMicros           = gSamplePeriodMicros[ channelIndex ];
			// TODO: for DC_TRIGGER inputs only, it might be useful to expose `t->inverted` here, via the appropriate menu item and global setting...
			t->threshold                      = gThresholds[ channelIndex ];
			t->minimumPulseDurationInMicros   = gDurationMinMicros[ channelIndex ];
			t->pulseDurationIncrementInMicros = gDurationIncMicros[ channelIndex ];
			t->refractoryPeriodInMillis       = gRefractoryMillis[ channelIndex ];
			t->mOutputPin1 = gUseSerialTriggerOutput ? NO_OUTPUT_PIN : gOutput1Pins[ channelIndex ];
			t->mOutputPin2 = gOutput2Pins[ channelIndex ];
		}
	}
}

void ConfigureSerialTriggerOutput( bool runningState )
{
	gUseSerialTriggerOutput = runningState;
	if( gUseSerialTriggerOutput ) gSerialTriggerOutput.begin( 115200 );
	else gSerialTriggerOutput.end();
	genie.bar( 0 );
	SetTriggerParameters();
}

void InitializeOutputPins( int pin1, int pin2, int outputLevel )
{
	if( pin1 >= 0 ) { pinMode( pin1, OUTPUT ); digitalWrite( pin1, outputLevel ); }
	if( pin2 >= 0 ) { pinMode( pin2, OUTPUT ); digitalWrite( pin2, outputLevel ); }
}

void InitializeOutputPins( ChannelIndex channelIndex )
{
	InitializeOutputPins( gOutput1Pins[ channelIndex ], gOutput2Pins[ channelIndex ], LOW );
}

void ConfigureTRS( JackIndex jackIndex, JackMode jackMode=KEEP_MODE, CircuitType circuitType=KEEP_CIRCUIT_TYPE )
{
	if( jackIndex < 0 || jackIndex > 3 ) return;
	int tipChannelIndex = jackIndex * 2;
	int ringChannelIndex = tipChannelIndex + 1;
	for( int i = tipChannelIndex; i <= ringChannelIndex; i++ )
	{
		switch( gChannelModes[ i ] )
		{
			case SWITCH:     delete ( Switch    * )gTriggers[ i ]; break;
			case AC_TRIGGER: delete ( Trigger   * )gTriggers[ i ]; break;
			case DC_TRIGGER: delete ( TriggerDC * )gTriggers[ i ]; break;
			default: break;
		}
		gTriggers[ i ] = NULL;
	}
	
	if( jackMode == KEEP_MODE ) jackMode = gJackModes[ jackIndex ];
	if( circuitType == KEEP_CIRCUIT_TYPE ) circuitType = gCircuitTypes[ jackIndex ];
	int readPin = NO_PIN, referencePin = NO_PIN, referenceVoltage = HIGH;
	if( circuitType == RING_HI_READ_TIP || circuitType == TIP_HI_READ_RING ) referenceVoltage = HIGH;
	if( circuitType == RING_LO_READ_TIP || circuitType == TIP_LO_READ_RING ) referenceVoltage = LOW;
	if( circuitType == RING_LO_READ_TIP || circuitType == RING_HI_READ_TIP ) { readPin = gInputPins[ tipChannelIndex  ]; referencePin = gInputPins[ ringChannelIndex ]; }
	if( circuitType == TIP_LO_READ_RING || circuitType == TIP_HI_READ_RING ) { readPin = gInputPins[ ringChannelIndex ]; referencePin = gInputPins[ tipChannelIndex  ]; }
	switch( jackMode )
	{
		case AC_TRIGGER:
		case DC_TRIGGER:
			if( circuitType == READ_BOTH || circuitType == RING_HI_READ_TIP || circuitType == RING_LO_READ_TIP )
				gTriggers[ tipChannelIndex  ] = ( jackMode == AC_TRIGGER ) ? ( void * )new   Trigger( gInputPins[ tipChannelIndex  ], gOutput1Pins[ tipChannelIndex  ] )
				                                                           : ( void * )new TriggerDC( gInputPins[ tipChannelIndex  ], gOutput1Pins[ tipChannelIndex  ] );

			if( circuitType == READ_BOTH || circuitType == TIP_HI_READ_RING || circuitType == TIP_LO_READ_RING )
				gTriggers[ ringChannelIndex ] = ( jackMode == AC_TRIGGER ) ? ( void * )new   Trigger( gInputPins[ ringChannelIndex ], gOutput1Pins[ ringChannelIndex ] )
				                                                           : ( void * )new TriggerDC( gInputPins[ ringChannelIndex ], gOutput1Pins[ ringChannelIndex ] );
			
			if( referencePin >= 0 ) { pinMode( referencePin, OUTPUT ); digitalWrite( referencePin, referenceVoltage ); }
			break;
			
		case SWITCH:		
			if( circuitType == READ_BOTH )
			{
				gTriggers[ tipChannelIndex  ] = ( void * )new Switch( gInputPins[ tipChannelIndex  ] );
				gTriggers[ ringChannelIndex ] = ( void * )new Switch( gInputPins[ ringChannelIndex ] );
			}
			else if( circuitType == RING_HI_READ_TIP || circuitType == RING_LO_READ_TIP )
				gTriggers[ tipChannelIndex  ] = ( void * )new Switch( readPin, referenceVoltage, referencePin, true );
			
			else if( circuitType == TIP_HI_READ_RING || circuitType == TIP_LO_READ_RING )
				gTriggers[ ringChannelIndex ] = ( void * )new Switch( readPin, referenceVoltage, referencePin, true );
			
			break;
			
		default:
			break;
	}
	gCircuitTypes[ jackIndex ] = circuitType;
	gJackModes[ jackIndex ] = jackMode;
	gChannelModes[  tipChannelIndex ] = gTriggers[  tipChannelIndex ] ? jackMode : DISABLED;
	gChannelModes[ ringChannelIndex ] = gTriggers[ ringChannelIndex ] ? jackMode : DISABLED;
	InitializeOutputPins( ( ChannelIndex )tipChannelIndex  );
	InitializeOutputPins( ( ChannelIndex )ringChannelIndex );
	SetTriggerParameters();
}

void IncludeInMenu( JackIndex jackIndex, bool setting=true )
{
	gVisibleInMenu[ jackIndex ] = setting;
}

int gThresholdDelta = 1;
unsigned long gDurationDeltaMicros = 10000;
unsigned long gRefractoryDeltaMillis = 10;

#define VALUEPTR( IND, FIELD )  ( !gTriggers[ IND ] ? NULL : ( gChannelModes[ IND ] == AC_TRIGGER || gChannelModes[ IND ] == DC_TRIGGER ) ? &( ( Trigger * )gTriggers[ IND ] )->FIELD : NULL )

void RenderNumericParameter( const char * valueFmt, double value, int millideltas=0 )
{
	static char * buffer = NULL;
	if( !buffer ) buffer = new char[ genie.LCD_COLS + 1 ];
	size_t valueLength = snprintf( buffer, genie.LCD_COLS + 1, valueFmt, value );
	size_t spaceLeft = ( valueLength < ( size_t )genie.LCD_COLS ) ? ( size_t )genie.LCD_COLS - valueLength : 0;
	String deltaString;
	switch( millideltas )
	{
		case       0: deltaString =          ""; break;
		case       1: deltaString =    "+0.001"; if( spaceLeft < deltaString.length() ) deltaString = "+1e-3"; break;
		case      10: deltaString =    "+0.010"; if( spaceLeft < deltaString.length() ) deltaString = "+0.01"; break;
		case     100: deltaString =    "+0.100"; if( spaceLeft < deltaString.length() ) deltaString = "+0.1";  break;
		case    1000: deltaString =    "+1";     break;
		case   10000: deltaString =   "+10";     break;
		case  100000: deltaString =  "+100";     break;
		case 1000000: deltaString = "+1000";     if( spaceLeft < deltaString.length() ) deltaString = "+1e3";  break;
	}
	if( spaceLeft < deltaString.length() && deltaString[ 1 ] == '0' ) deltaString = "+" + deltaString.substring( 2 );
	if( spaceLeft < deltaString.length() ) deltaString = deltaString.substring( 1 );
	if( deltaString[ 0 ] == '+' ) deltaString[ 0 ] = '\x01';
	for( ; spaceLeft > deltaString.length(); spaceLeft-- ) sprintf( buffer + valueLength++, " " );
	if( spaceLeft >= deltaString.length() ) sprintf( buffer + valueLength, "%s", deltaString.c_str() );
	genie.lcd.print( buffer );
}

void RenderThreshold( ChannelIndex channelIndex )
{
	int * valuePtr = VALUEPTR( channelIndex, threshold );
	RenderNumericParameter( "%g", valuePtr ? ( double )*valuePtr : -1.0, gThresholdDelta * 1000 );
}
void RenderMinLength( ChannelIndex channelIndex )
{
	unsigned long * valuePtr = VALUEPTR( channelIndex, minimumPulseDurationInMicros );
	RenderNumericParameter( "%5.3f", valuePtr ? ( double )*valuePtr / 1000.0 : -1.0, gDurationDeltaMicros );
}
void RenderIncrement( ChannelIndex channelIndex )
{
	unsigned long * valuePtr = VALUEPTR( channelIndex, pulseDurationIncrementInMicros );
	RenderNumericParameter( "%5.3f", valuePtr ? ( double )*valuePtr / 1000.0 : -1.0, gDurationDeltaMicros );
}
void RenderRefractMs( ChannelIndex channelIndex )
{
	unsigned long * valuePtr = VALUEPTR( channelIndex, refractoryPeriodInMillis );
	RenderNumericParameter( "%g", valuePtr ? ( double )*valuePtr : -1.0, gRefractoryDeltaMillis * 1000 );
}
void RenderSampleRate( JackIndex jackIndex=ALL_JACKS )
{
	unsigned long * valuePtr = NULL;
	for( int channelIndex = 0; channelIndex < NUMBER_OF_CHANNELS; channelIndex++ )
	{
		if( jackIndex != ALL_JACKS && channelIndex != jackIndex * 2 && channelIndex != jackIndex * 2 + 1 ) continue;
		valuePtr = VALUEPTR( channelIndex, samplePeriodInMicros );
		if( valuePtr ) break;
	}
	unsigned long samplesPerSecond = ( valuePtr && *valuePtr ) ? ( 1000000 / *valuePtr ) : -1;
	RenderNumericParameter( "%g Hz", samplesPerSecond, 0 );
}

void RenderLoopsPerBar()
{
	RenderNumericParameter( "%g", gLoopsPerBar, 0 );
}

bool RenderTemperature( bool modified=false )
{
	if( modified ) return false;
	RenderNumericParameter( "%.1f\x02" "C turn=redo", genie.temperature() );
	return true;
}

bool AdjustThreshold( int direction, ChannelIndex channelIndex, bool modified )
{
	if( modified )
	{
#		if MODIFIED_LEFT_ALWAYS_GOES_BACK
			if( direction < 0 ) return false;
			if( gThresholdDelta >= 100 ) gThresholdDelta = 1;
			else gThresholdDelta *= 10;
#		else
			if( direction > 0 && gThresholdDelta < 100 ) gThresholdDelta *= 10;
			if( direction < 0 && gThresholdDelta >= 10 ) gThresholdDelta /= 10;
#		endif
	}
	else
	{
		if( direction > 0 ) gThresholds[ channelIndex ] += gThresholdDelta;
		if( direction < 0 ) gThresholds[ channelIndex ] = ( gThresholds[ channelIndex ] > gThresholdDelta ) ? gThresholds[ channelIndex ] - gThresholdDelta : 0;
		SetTriggerParameters();
	}
	return true;
}
bool AdjustMinLength( int direction, ChannelIndex channelIndex, bool modified )
{
	if( modified )
	{
#		if MODIFIED_LEFT_ALWAYS_GOES_BACK
			if( direction < 0 ) return false;
			if( gDurationDeltaMicros >= 1000000 ) gDurationDeltaMicros = 1;
			else gDurationDeltaMicros *= 10;
#		else
			if( direction > 0 && gDurationDeltaMicros < 1000000 ) gDurationDeltaMicros *= 10;
			if( direction < 0 && gDurationDeltaMicros >=     10 ) gDurationDeltaMicros /= 10;
#		endif
	}
	else
	{
		if( direction > 0 ) gDurationMinMicros[ channelIndex ] += gDurationDeltaMicros;
		if( direction < 0 ) gDurationMinMicros[ channelIndex ] = ( gDurationMinMicros[ channelIndex ] > gDurationDeltaMicros ) ? gDurationMinMicros[ channelIndex ] - gDurationDeltaMicros : 0;
		SetTriggerParameters();
	}
	return true;
}
bool AdjustIncrement( int direction, ChannelIndex channelIndex, bool modified )
{
	if( modified )
	{
#		if MODIFIED_LEFT_ALWAYS_GOES_BACK
			if( direction < 0 ) return false;
			if( gDurationDeltaMicros >= 1000000 ) gDurationDeltaMicros = 1;
			else gDurationDeltaMicros *= 10;
#		else
			if( direction > 0 && gDurationDeltaMicros < 1000000 ) gDurationDeltaMicros *= 10;
			if( direction < 0 && gDurationDeltaMicros >=     10 ) gDurationDeltaMicros /= 10;
#		endif
	}
	else
	{
		if( direction > 0 ) gDurationIncMicros[ channelIndex ] += gDurationDeltaMicros;
		if( direction < 0 ) gDurationIncMicros[ channelIndex ] = ( gDurationIncMicros[ channelIndex ] > gDurationDeltaMicros ) ? gDurationIncMicros[ channelIndex ] - gDurationDeltaMicros : 0;
		SetTriggerParameters();
	}
	return true;
}
bool AdjustRefractMs( int direction, ChannelIndex channelIndex, bool modified )
{
	if( modified )
	{
#		if MODIFIED_LEFT_ALWAYS_GOES_BACK
			if( direction < 0 ) return false;
			if( gRefractoryDeltaMillis >= 1000 ) gRefractoryDeltaMillis = 1;
			else gRefractoryDeltaMillis *= 10;
#		else
			if( direction > 0 && gRefractoryDeltaMillis <  1000 ) gRefractoryDeltaMillis *= 10;
			if( direction < 0 && gRefractoryDeltaMillis >=   10 ) gRefractoryDeltaMillis /= 10;
#		endif
	}
	else
	{
		if( direction > 0 ) gRefractoryMillis[ channelIndex ] += gRefractoryDeltaMillis;
		if( direction < 0 ) gRefractoryMillis[ channelIndex ] = ( gRefractoryMillis[ channelIndex ] > gRefractoryDeltaMillis ) ? gRefractoryMillis[ channelIndex ] - gRefractoryDeltaMillis : 0;
		SetTriggerParameters();
	}
	return true;
}
bool AdjustSampleRate( int direction, JackIndex jackIndex, bool modified )
{
#	if MODIFIED_LEFT_ALWAYS_GOES_BACK
		if( modified ) return ( direction < 0 ) ? false : true; // press-and-turn-RIGHT is supposed to adjust the delta, but there is no delta, so that does nothing; however press-and-turn-LEFT performs the default action
#	else
		if( modified ) return true; // press-and-turn in either direction is supposed to adjust the delta, but there is no delta, so it does nothing
#	endif
	direction *= -1; /* increasing sample rate means decreasing gSamplePeriodMicros */
	for( int channelIndex = 0; channelIndex < NUMBER_OF_CHANNELS; channelIndex++ )
	{
		if( jackIndex != ALL_JACKS && channelIndex != jackIndex * 2 && channelIndex != jackIndex * 2 + 1 ) continue;
		unsigned long & period = gSamplePeriodMicros[ channelIndex ];
		do
		{
			if( direction <= 0 && period <=   20 ) { period =   20; break; } //   20 micros <=> 50 kHz
			if( direction >= 0 && period >= 2000 ) { period = 2000; break; } // 2000 micros <=>   500 Hz
			period += direction; 
		} while( direction && 1000000 % period );
	}
	SetTriggerParameters();
	return true;
}
bool AdjustLoopsPerBar( int direction, bool modified )
{
#	if MODIFIED_LEFT_ALWAYS_GOES_BACK
		if( modified ) return ( direction < 0 ) ? false : true; // press-and-turn-RIGHT is supposed to adjust the delta, but there is no delta, so that does nothing; however press-and-turn-LEFT performs the default action
#	else
		if( modified ) return true; // press-and-turn in either direction is supposed to adjust the delta, but there is no delta, so it does nothing
#	endif
	if( direction < 0 && gLoopsPerBar > 1 ) gLoopsPerBar--;
	else if( direction > 0 ) gLoopsPerBar++;
	return true;
}

//#define SETTINGS_SDCARD BUILTIN_SDCARD  // leave this line commented-out to use EEPROM; uncomment it to use the SD card (this definition must happen before #include "PersistentSettings.h")
#include "PersistentSettings.h"
bool Settings( bool writing ) // load or save settings
{
	OPEN_SETTINGS( "FourJacksMenuDriven-V6", writing ); // the string should be specific to the order, data formats and roles of the settings that follow
	bool success = true;
	for( int i = 0; i < NUMBER_OF_CHANNELS; i++ )
	{
		success = success && SETTING( gInputPins[ i ] );
		success = success && SETTING( gOutput1Pins[ i ] );
		success = success && SETTING( gOutput2Pins[ i ] );
		success = success && SETTING( gThresholds[ i ] );
		success = success && SETTING( gDurationMinMicros[ i ] );
		success = success && SETTING( gDurationIncMicros[ i ] );
		success = success && SETTING( gRefractoryMillis[ i ] );
		success = success && SETTING( gSamplePeriodMicros[ i ] );
		success = success && SETTING( gChannelModes[ i ] );
	}
	for( int i = 0; i < NUMBER_OF_JACKS; i++ )
	{
		success = success && SETTING( gJackModes[ i ] );
		success = success && SETTING( gCircuitTypes[ i ] );
		// don't include gVisibleInMenu - we want that controlled from inside the sketch
	}
	success = success && SETTING( gUseSerialTriggerOutput );
	
	if( writing ) return success;
	for( int i = 0; i < NUMBER_OF_JACKS; i++ ) ConfigureTRS( ( JackIndex )i );
	return success;
}

bool AvailableForBNC( int bncPin, ChannelIndex channelIndex )
{	// the channel is available if it is not disabled and doesn't already have a different BNC assigned to it
	if( gChannelModes[ channelIndex ] == DISABLED ) return false;
	if( gOutput2Pins[ channelIndex ] >= 0 && gOutput2Pins[ channelIndex ] != bncPin ) return false;
	return true;
}

void AssignBNC( int bncPin, ChannelIndex channelIndex )
{
	for( int i = 0; i < NUMBER_OF_CHANNELS; i++ )
	{
		if( gOutput2Pins[ i ] == bncPin ) gOutput2Pins[ i ] = NO_OUTPUT_PIN;
	}
	if( channelIndex != NO_CHANNEL )
	{
		gOutput2Pins[ channelIndex ] = bncPin;
		InitializeOutputPins( channelIndex );
	}
	SetTriggerParameters();
}

ChannelIndex FindBNC( int bncPin )
{
	if( bncPin < 0 ) return NO_CHANNEL;
	for( int channelIndex = 0; channelIndex < NUMBER_OF_CHANNELS; channelIndex++ )
		if( gOutput2Pins[ channelIndex ] == bncPin ) return ( ChannelIndex )channelIndex;
	return NO_CHANNEL;
}

#include "EightChannelDemoMenu.h" // auto-generated by MakeMenu.py; contains code that references the global definitions above
Menu menu( genie.lcd );

void setup( void )
{
	genie.begin();
	InitializeOutputPins( genie.BNC1, genie.BNC2, LOW );
	
	if( !Settings( LOAD_SETTINGS ) ) // may fail and return early if no settings have been previously saved. 
		for( int i = 0; i < NUMBER_OF_JACKS; i++ ) ConfigureTRS( ( JackIndex )i, AC_TRIGGER );
	
	for( int i = 0; i < NUMBER_OF_JACKS; i++ ) IncludeInMenu( ( JackIndex ) i );
	menu.begin();

	byte delta[ 8 ] = {
		0b00000,
		0b00000,
		0b00100,
		0b01010,
		0b01010,
		0b10001,
		0b11111,
		0b00000,
	};
	genie.lcd.createChar( 1, delta );
	byte degree[ 8 ] = {
		0b00100,
		0b01010,
		0b00100,
		0b00000,
		0b00000,
		0b00000,
		0b00000,
		0b00000,
	};
	genie.lcd.createChar( 2, degree );
}

#ifdef KEYHOLE
	KEYHOLE keyhole;
#endif
void loop( void )
{
	Run();
	int selectorWheelEvent  = genie.selectorWheel.event();
	int selectorButtonEvent = genie.selectorButton.event();
#	ifdef KEYHOLE
		if( keyhole.begin() )
		{
			if( keyhole.command( "h" ) ) menu.home(); 
			if( keyhole.command( "l" ) ) selectorWheelEvent  = Wheel::LEFT;
			if( keyhole.command( "r" ) ) selectorWheelEvent  = Wheel::RIGHT;
			if( keyhole.command( "s" ) ) selectorButtonEvent = Switch::RELEASE;
			if( keyhole.command( "x" ) )
			{
				menu.modifyNavigation = !menu.modifyNavigation;
				if( !menu.modifyNavigation ) selectorButtonEvent = Switch::RELEASE;
			}
			keyhole.end();
		}
#	endif
	if( selectorWheelEvent == Wheel::LEFT  ) menu.left();
	if( selectorWheelEvent == Wheel::RIGHT ) menu.right();
	if( selectorWheelEvent && menu.modifyNavigation ) menu.ignoreNextSelection();
	if( selectorButtonEvent == Switch::PRESS   ) menu.modifyNavigation = true;
	if( selectorButtonEvent == Switch::RELEASE ) { menu.modifyNavigation = false; menu.select(); }
	if( genie.pushButton.event() == Switch::RELEASE ) menu.home();  // the top-side push-button takes you home()
}
