/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/
/*
This sketch demonstrates two functions of the SyncGenie library: how to
construct custom on-screen menus, and how to save and load settings. The menu
system is defined in the `DEMO_SPEC` of `tools/MakeMenu.py` (which gets used if
`MakeMenu.py` is run without any input arguments). When you run `MakeMenu.py`
in Python, it will spit out the contents of a header file, which this sketch
expects to #include as "TestMenu.h".

The menus present an interface to a fake surround-sound audio setup, allowing
you to change the imaginary speaker configuration from 2.0 (stereo) to 5.1 or
7.1 surround, and then adjust the relative volumes of the speakers that are
available under that configuration. A "SAVE NOW" option is included under the
"Stored Settings" menu: when you select this, the config and volumes will be
saved and will survive turning the SyncGenie off and on again.
*/

#include "SyncGenie.h"
SyncGenie genie;

///////////////////////////////////////////////////////////////////////////
// Define/declare global variables and functions that the menus will use.

double gConfig = 2.0;
unsigned short gFront = 50, gSurround = 50, gBackSurround = 50, gCenter = 50, gSubwoofer = 50;

// In our `DEMO_SPEC` code, this function is called when the user
// turns the selector to adjust a numeric setting:
bool Adjust( int direction, unsigned short & level, bool modified )
{
	// Our `DEMO_SPEC` menu code passes the Menu's `modifyNavigation`
	// member as the third argument to this function. Which means,
	// `modified=true` indicates the selector is pressed-in while turning.
	// In normal menu navigation, that's a shortcut for going up or
	// down the tree. However, here in this custom hook, we can choose
	// how to interpret it. Let's use it for finer-grained adjustment:
	unsigned short delta = modified ? 1 : 10;
	if( direction < 0 ) level = ( level < delta       ) ?   0 : level - delta;
	if( direction > 0 ) level = ( level > 100 - delta ) ? 100 : level + delta; 
	return true; // Our `DEMO_SPEC` code passes this return value out to the 
	// `MenuBase` framework, where a `true` value is interpreted to mean,
	// "I have handled this event, so do not perform the default navigation
	// action"
}


//#define SETTINGS_SDCARD BUILTIN_SDCARD  // leave this line commented-out to use EEPROM; uncomment it to use the SD card (this definition must happen before #include "PersistentSettings.h")
#include "PersistentSettings.h"   // this header defines the OPEN_SETTINGS() and SETTING() macros, and the constants LOAD_SETTINGS=0 and SAVE_SETTINGS=1

// In our `DEMO_SPEC` code, this function is called when the user selects 
// either "SAVE NOW" or "REVERT TO SAVED" from the "Stored Settings" menu:
bool Settings( bool areWeSaving )
{
	OPEN_SETTINGS( "FakeAudioConfigDemo-V1", areWeSaving );
	
	// The "filename" in the line above should be specific to the order, data
	// formats and roles of the settings that follow: if you change, add or
	// remove any `SETTING()`, or even change the data type of any of the
	// variables involved, then you should at least bump the "-V1" version
	// suffix to a higher number.
	// 
	// If you have #defined `SETTINGS_SDCARD` (above), and assuming you have
	// inserted a working formatted SD card into the Teensy, this string will
	// literally be used as a filename. If not, the EEPROM can be thought of
	// as a file-system that has room for only one file: if that file's "name"
	// matches the expected string for the current sketch/format, the saved data
	// will be considered valid, and `Settings(false)` will load them; if not,
	// the settings will be considered absent, and `Settings(false)` will have
	// no effect (sketch variables will stay at their default/current values).
	// `Settings(true)` will of course overwrite the file---and on the EEPROM,
	// that also means it gets "renamed".
	// 
	// The Teensy's flash-emulated EEPROM can only be written about 10,000 times
	// before the flash memory units start to fail. So, `Settings(true)` should
	// not be called in any kind of automated loop---it should only be triggered
	// manually (and if you save your settings once per day for longer than 27
	// years, you might get problems).
	
	bool success = true;
	success = success && SETTING( gConfig );
	success = success && SETTING( gFront );
	success = success && SETTING( gSurround );
	success = success && SETTING( gBackSurround );
	success = success && SETTING( gCenter );
	success = success && SETTING( gSubwoofer );
	return success;
}

///////////////////////////////////////////////////////////////////////////

#include "TestMenu.h" // this header file gets auto-generated by MakeMenu.py; it defines the `Menu` class, which contains code that references the global definitions above
Menu menu( genie.lcd ); // instantiate the Menu

///////////////////////////////////////////////////////////////////////////

void setup( void )
{
	genie.begin(); // this initializes some pin modes, starts the LCD screen, starts the Serial connection if it's available, and flickers D9_LED (same as LED_BUILTIN) to say hello
	Settings( LOAD_SETTINGS ); // if settings specific to this sketch have been previously saved, this loads them now
	menu.begin();  // this enables the menu system, starting us off at the root menu node
}

void loop( void )
{
	int selectorWheelEvent  = genie.selectorWheel.event();  // will be zero unless the selector has been turned left or right since the previous call
	int selectorButtonEvent = genie.selectorButton.event(); // will be zero unless the selector has been pressed or released since the previous call
	int pushButtonEvent     = genie.pushButton.event();     // will be zero unless the top-side push-button has been pressed or released since the previous call

	// Handle selector press and release events:
	if( selectorButtonEvent == Switch::PRESS   ) menu.modifyNavigation = true;
	if( selectorButtonEvent == Switch::RELEASE ) { menu.modifyNavigation = false; menu.select(); } // selection actually happens when you release the selector
	
	// If you're pressing-and-turning, let's ensure your subsequent release of the selector is not interpreted as an additional "click":
	if( selectorWheelEvent && menu.modifyNavigation ) menu.ignoreNextSelection();
	
	// Handle wheel-turn events:
	if( selectorWheelEvent == Wheel::LEFT     ) menu.left();
	if( selectorWheelEvent == Wheel::RIGHT    ) menu.right();
	
	// Press the top-side push-button to go back to the start of the menu:
	if( pushButtonEvent    == Switch::RELEASE ) menu.home(); 
	// (You do not have to do this, of course. You can reserve the top-side push-button for other functions instead. Going 
	// to the top of the menu is easy even without a dedicated button: just press & hold while turning left a few notches.)
}
