/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
This sketch is intended not so much as a programming tutorial, but rather as
a utility for testing connectivity by controlling of the output voltages on
the SyncGenie's connectors, setting each pin `HIGH` or `LOW`. Use the serial
port to issue commands like `d5 = 1`  or  `t2 = 0`.

- `d0` through `d9` access `D0_OUT` through `D7_OUT`, respectively, followed
  by `D8_LED` and `D9_LED`.

- `i0` through `i7` access `D0_IN` through `D7_IN`, respectively, (which,
  although generally intended for input, should also work here as outputs).
   
- `t1` through `t4` access the tips of TRS1 through TRS4, respectively.

- `r1` through `r4` access the rings of TRS1 through TRS4, respectively.
   
- `b1` and `b2` access the BNCs.

- `j3` and `j4` access the third and fourth JST pins.

NB: you won't ever measure a `HIGH` output voltage on t4 and r4, unless you
    have shorted their respective bias circuits via their exposed board headers
    (TRS4 is intended only for input when the bias circuits are operational).
    
This sketch requires the `Keyhole` library, which you may have to install
by hand via the Arduino IDE's Library Manager.
*/

#include "SyncGenie.h"
SyncGenie genie;

#include "Keyhole.h" // you may have to install the Keyhole library by hand (we have not made it a dependency
KEYHOLE keyhole;     // of the SyncGenie library, even though it is needed in some of the example sketches)

void setup( void )
{
	genie.begin( "Utility: Pin\nOutput Control" );	
}

void loop( void )
{
	if( keyhole.begin() )
	{
#		define CONTROL( VAR, PIN ) static bool VAR; if( keyhole.variable( #VAR, VAR ) ) { pinMode( PIN, OUTPUT ); digitalWrite( PIN, VAR ? HIGH : LOW ); }
		CONTROL( d0, genie.D0_OUT );
		CONTROL( d1, genie.D1_OUT );
		CONTROL( d2, genie.D2_OUT );
		CONTROL( d3, genie.D3_OUT );
		CONTROL( d4, genie.D4_OUT );
		CONTROL( d5, genie.D5_OUT );
		CONTROL( d6, genie.D6_OUT );
		CONTROL( d7, genie.D7_OUT );
		CONTROL( d8, genie.D8_LED );
		CONTROL( d9, genie.D9_LED );
		CONTROL( t1, genie.TRS1_TIP  );
		CONTROL( r1, genie.TRS1_RING );
		CONTROL( t2, genie.TRS2_TIP  );
		CONTROL( r2, genie.TRS2_RING );
		CONTROL( t3, genie.TRS3_TIP  );
		CONTROL( r3, genie.TRS3_RING );
		CONTROL( t4, genie.TRS4_TIP  ); // you may never be able to observe HIGH output on TRS4_TIP due to its bias circuit. It is intended only for input when the bias circuit is operational.
		CONTROL( r4, genie.TRS4_RING ); // you may never be able to observe HIGH output on TRS4_RING due to its bias circuit. It is intended only for input when the bias circuit is operational.
		CONTROL( b1, genie.BNC1 );
		CONTROL( b2, genie.BNC2 );
		CONTROL( j3, genie.JST_PIN3 );
		CONTROL( j4, genie.JST_PIN4 );
		CONTROL( i0, genie.D0_IN );
		CONTROL( i1, genie.D1_IN );
		CONTROL( i2, genie.D2_IN );
		CONTROL( i3, genie.D3_IN );
		CONTROL( i4, genie.D4_IN );
		CONTROL( i5, genie.D5_IN );
		CONTROL( i6, genie.D6_IN );
		CONTROL( i7, genie.D7_IN );
		
		keyhole.end();
	}
}
