/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
This sketch illustrates how to use the SerialTriggerOutput class, to accomodate
acquisition devices that accept triggers as serial messages rather than raw TTL
pulses (for example, the FlexEEG by NeuroCONCISE). It uses one standard `Trigger`
(audio signal on TRS1_TIP) and two push-buttons (the top-side button and the
selector button) to form three independent trigger bits. (NB: the FlexEEG can
record up to 6 bits; the `SerialTriggerOutput` class itself can handle up to 8
bits). When connecting to the FlexEEG, we use the SyncGenie's `D6_OUT` pin
(pin 8 of the female D-sub25 output) and connect this via an optical isolator
(built into the cable connector). You will see the D6 LED light up on the
bar graph, and none of the others (actually it is flickering very fast, at
115200 baud).
*/

#include <SyncGenie.h>
SyncGenie genie;

#include <Triggering.h>
Trigger trigger( genie.TRS1_TIP, NO_OUTPUT_PIN );

#include <SerialTriggerOutput.h>
SerialTriggerOutput flexX;


void setup()
{
	genie.begin( "Demo sketch:\nSerial Output" ); // this initializes some pin modes, starts the LCD screen, starts the Serial connection, and flickers D9_LED (same as LED_BUILTIN) to say hello

	flexX.begin(); // must do this *after* `genie.begin()`
	//flexX.debug = "sent"; // TODO: for best performance, comment this out - it causes trigger messages to be echoed as text to the main (USB) serial port
}

void loop()
{
	// Gather some inputs:

	unsigned long now = micros();
	
	trigger.process( now );
	
	static bool pbPressed = false;
	switch( genie.pushButton.event() ) // you could just say `pbPressed = genie.pushButton.isClosed();` but the `event()` method has the advantage of debouncing (i.e. ignoring very short releases)
	{
		case Switch::PRESS:   pbPressed = true;  break;
		case Switch::RELEASE: pbPressed = false; break;
		default: break;
	}

	static bool selectorPressed = false;
	switch( genie.selectorButton.event() ) // you could just say `selectorPressed = genie.selectorButton.isClosed();` but the `event()` method has the advantage of debouncing (i.e. ignoring very short releases)
	{
		case Switch::PRESS:   selectorPressed = true;  break;
		case Switch::RELEASE: selectorPressed = false; break;
		default: break;
	}


	// Handle the trigger output:

	flexX.reset(); // set the trigger byte to zero
	flexX.setBit( 0, trigger );         // lowest bit is the audio trigger pulse
	flexX.setBit( 1, pbPressed );       // second bit goes high when the top-side push-button is pressed
	flexX.setBit( 2, selectorPressed ); // third bit goes high when the selector button is pressed

	flexX.send();  // if the value of the trigger byte has changed since the last `.send()`, and it is now non-zero, its value will be sent.
	               // NB: for the FlexEEG headset there's no point in sending an explicit 0 when the byte changes to 0, because the headset
	               // only records events as transients anyway; however, if you're working with a different device and need to announce
	               // changes-to-zero explicitly, you can say `.send(true)`.
}
