/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
This example sketch sets up a `TriggerDC` instance that reads a light-sensor
signal on `TRS1_TIP` and generates pulses on `D0_OUT`. You can inspect and
interact with the `TriggerDC` instance's parameters via the Arduino IDE's
Serial Monitor window, courtesy of the Keyhole library (NB: you may have to
install Keyhole by hand in the IDE's Library Manager; we have not made Keyhole
an automatic dependency of the SyncGenie library, since it is not required for
library functionality - only for some of the example sketches where serial-port
interaction is needed).

The light sensor (whether it is a light-dependent resistor or phototransistor)
will also need a supply voltage, which we will supply on `TRS1_RING`.
NB: you will probably find you need a much higher threshold value for an LDR
than for a photo-transistor.
*/

#include <SyncGenie.h>
SyncGenie genie;

const int lightSignalPin       = genie.TRS1_TIP;  // Let's read a light sensor (phototransistor) on the tip of TRS1.
const int lightSensorSupplyPin = genie.TRS1_RING; // We'll need to supply voltage to the photo-transistor, so let's do that on the ring of TRS1. 
#include <Triggering.h>
TriggerDC trigger( lightSignalPin, genie.D0_OUT );

#include "Keyhole.h" // you may have to install the Keyhole library by hand (we have not made it a dependency
KEYHOLE keyhole;     // of the SyncGenie library, even though it is needed in some of the example sketches)


void setup()
{
	genie.begin( "Demo sketch:\nTriggerDC Params" ); // this initializes some pin modes, starts the LCD screen, starts the Serial connection, and flickers D9_LED (same as LED_BUILTIN) to say hello
	
	trigger.rememberSamples( 2.0 ); // makes the object calculate .recentMax in a 2-second sliding window
	keyhole.autoSeconds = 2.0; // by default, report variables to the Serial console every 2 seconds

	// Supply voltage to the photo-sensor:
	pinMode( lightSensorSupplyPin, OUTPUT ); // genie.begin() does not set pin modes on any TRS, BNC or JST pins, so we have to do that manually
	digitalWrite( lightSensorSupplyPin, HIGH );
}

void loop()
{
	unsigned long now = micros();
	
	trigger.process( now );
		
	if( keyhole.begin( now ) )
	{
		// Expose the Trigger instance's various parameters. Watch how `recent_max`
		// changes according to the level of the input you're feeding in on TRS1_TIP. Use
		// this to inform your choice of an appropriate value for `threshold` (send a
		// command like `threshold=30` to configure it).
		if( keyhole.variable( "micros_per_sample", trigger.samplePeriodInMicros ) ) // if sampling frequency changes,
			trigger.rememberSamples( 2.0 );                                         // the "recent max" memory needs to be reorganized
		keyhole.variable( "recent_max", trigger.recentMax, VARIABLE_READ_ONLY );
		keyhole.variable( "threshold", trigger.threshold );
		keyhole.variable( "min_micros", trigger.minimumPulseDurationInMicros );
		keyhole.variable( "inc_micros", trigger.pulseDurationIncrementInMicros );
		keyhole.variable( "refrac_millis", trigger.refractoryPeriodInMillis );		
				
		// A command like `auto_seconds = 2.0` causes all exposed variables to be reported
		// to the serial console automatically every two seconds (or however many seconds
		// you specify). Set it <= 0 to turn auto-reporting off (you can still query the
		// variables manually by sending `?`).
		keyhole.variable( "auto_seconds", keyhole.autoSeconds );
		
		keyhole.end();
	}
}
