/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/
/*
This file is for debugging `genie.lcd` when you do not have LCD hardware attached
to your microcontroller. To use it, create a copy of this file, here in the same
`src` directory, and call it `LiquidCrystal.h`.  This will overshadow the board
distro's header of the same name and the associated implementation. Delete that
copy when you are done. Be careful not to add it inadvertently to version control
(but do not .gitignore it, lest you forget that it is there). 
*/

#ifndef   __FakeLiquidCrystal_H__
#define   __FakeLiquidCrystal_H__
#include <Arduino.h>

#define DBSTREAM Serial // activates the REPORT() and REPORTS() macros in Keyhole.h
#include "Keyhole.h" // you'll have to install this by hand.  It's not a dependency of the SyncGenie library as a whole

class LiquidCrystal
{
	public:
		LiquidCrystal( uint8_t, uint8_t, uint8_t, uint8_t, uint8_t, uint8_t ) : mColumn( 0 ), mRow( 0 ), mNumberOfColumns( 0 ), mNumberOfRows( 0 ), mLines( NULL ) { }
		~LiquidCrystal() { this->end(); }
	
		void begin( uint8_t nColumns, uint8_t nRows, uint8_t charSize=0 )
		{
			this->end();
			mNumberOfColumns = nColumns;
			mNumberOfRows = nRows;
			mLines = new String[ mNumberOfRows ];
		}
		void end( void )
		{
			mNumberOfColumns = mNumberOfRows = 0;
			delete [] mLines; mLines = NULL;
			mPreviousOutput = "";
		}
		void clear( void )
		{
			for( unsigned int i = 0; i < mNumberOfRows && mLines; i++ ) mLines[ i ] = "";
			this->setCursor( 0, 0 );
		}
		void setCursor( uint8_t column, uint8_t row )
		{
			mColumn = column;
			mRow = row;
		}
		void print( const unsigned short x ) { print( ( double )x ); }
		void print( const int            x ) { print( ( double )x ); }
		void print( const double x )
		{
			static const size_t bufferSize = 17;
			char buffer[ bufferSize ];
			snprintf( buffer, bufferSize, "%g", x );
			this->print( buffer );
		}
		void print( const char c )
		{
			if( !mLines || mRow >= mNumberOfRows ) return;
			String & line = mLines[ mRow ];
			if( line.length() > mColumn ) line = line.substring( 0, mColumn );
			while( line.length() < mColumn ) line += ' ';
			line += c;
			mColumn = line.length();
			if( mRow == mNumberOfRows - 1 && line.length() == mNumberOfColumns ) this->flush();
		}
		size_t write( uint8_t c )
		{
			print( c ); return 1;
		}
		void print( const String & s )
		{
			if( !mLines || mRow >= mNumberOfRows ) return;
			String & line = mLines[ mRow ];
			if( line.length() > mColumn ) line = line.substring( 0, mColumn );
			while( line.length() < mColumn ) line += ' ';
			line += s;
			mColumn = line.length();
			if( mRow == mNumberOfRows - 1 ) this->flush();
		}
		
		void flush( void )
		{
			if( !mLines ) { Serial.println( "## no LCD lines to display" ); return; }
			String output;
			for( unsigned int iRow = 0; iRow < mNumberOfRows; iRow++ )
			{
				output += '|';
				for( unsigned int iColumn = 0; iColumn < mNumberOfColumns; iColumn++ )
					output += ( iColumn < mLines[ iRow ].length() ) ? mLines[ iRow ][ iColumn ] : ' ';
			}
			output += '|';
			if( output != mPreviousOutput )
			{
				Serial.println( output );
				Serial.flush();
				mPreviousOutput = output;
			}
		}
		
		void createChar( uint8_t, const uint8_t[] ) {}
		
	private:
		unsigned int mColumn;
		unsigned int mRow;
		unsigned int mNumberOfColumns;
		unsigned int mNumberOfRows;
		String * mLines;
		String mPreviousOutput;
};
#endif // __FakeLiquidCrystal_H__
