/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __MenuBase_CPP__
#define   __MenuBase_CPP__

#include "MenuBase.h"
#include <Arduino.h> // for delay()

MenuBase::MenuBase( LiquidCrystal & _lcd, uint16_t nSlots, MenuLabel title )
: lcd( _lcd ), modifyNavigation( false ),
  mMaxNumberOfNodes( nSlots ), mNumberOfNodes( 0 ), mCurrentNode( 0 ),
  mIgnoreNextSelection( false ),
  mParentNode( NULL ), mString( NULL )
{
	mParentNode    = new uint16_t[        mMaxNumberOfNodes ];
	mString        = new StoredMenuLabel[ mMaxNumberOfNodes ];
	this->addNode( 0, title );
}

MenuBase::~MenuBase()
{
	delete [] mParentNode;    mParentNode    = NULL;
	delete [] mString;        mString        = NULL;
}
		
void 
MenuBase::begin( void )
{
	this->home();
	this->modifyNavigation = false;
	mIgnoreNextSelection = false;
}

uint16_t 
MenuBase::addNode( uint16_t parentNode, MenuLabel text )
{
	if( mNumberOfNodes >= mMaxNumberOfNodes ) return 0;
	mParentNode[ mNumberOfNodes ] = parentNode;
	mString[ mNumberOfNodes ] = text;
	if( mCurrentNode == 0 ) mCurrentNode = mNumberOfNodes;
	return mNumberOfNodes++;
}
		
void MenuBase::init( void ) { }
bool MenuBase::isVisible(    uint16_t nodeIndex ) { return true;  }
bool MenuBase::isSelected(   uint16_t nodeIndex ) { return false; }
bool MenuBase::customRender( uint16_t nodeIndex ) { return false; }
bool MenuBase::customLeft(   uint16_t nodeIndex ) { return false; }
bool MenuBase::customRight(  uint16_t nodeIndex ) { return false; }
bool MenuBase::customSelect( uint16_t nodeIndex ) { return false; }
		
void 
MenuBase::render( unsigned short nBlinks, unsigned long msecPerBlink )
{
	this->lcd.clear();
	this->lcd.print( mString[ mParentNode[ mCurrentNode ] ] );
#	ifdef __FakeLiquidCrystal_H__
		nBlinks = 0;
#	endif
	while( true )
	{
		this->lcd.setCursor( 0, 1 );
		if( !this->customRender( mCurrentNode ) )
			this->lcd.print( mString[ mCurrentNode ] );
		
		if( !nBlinks-- ) break;
		delay( msecPerBlink );
		this->lcd.setCursor( 0, 1 );
		for( int i = 0; i < 40; i++ ) this->lcd.print( " " );
		delay( msecPerBlink );
	}
}

void 
MenuBase::left( void )
{
	if( mCurrentNode == 0 || mNumberOfNodes <= 1 ) return;
	if( !this->customLeft( mCurrentNode ) )
	{
		if( this->modifyNavigation ) { this->back(); return; }
		uint16_t start = mCurrentNode, parent = mParentNode[ mCurrentNode ];
		do
		{
			if( --mCurrentNode == 0 ) mCurrentNode = mNumberOfNodes - 1;
			if( mParentNode[ mCurrentNode ] == parent && this->isVisible( mCurrentNode ) ) break;
		} while( mCurrentNode != start );
	}
	this->render();
}

void 
MenuBase::right( void )
{
	if( mCurrentNode == 0 || mNumberOfNodes <= 1 ) return;
	if( !this->customRight( mCurrentNode ) )
	{
		if( this->modifyNavigation )
		{
			bool ignoreNextSelection = mIgnoreNextSelection; mIgnoreNextSelection = false;
			this->select();
			mIgnoreNextSelection = ignoreNextSelection;
			return;
		}
		uint16_t start = mCurrentNode, parent = mParentNode[ mCurrentNode ];
		do
		{
			if( ++mCurrentNode >= mNumberOfNodes ) mCurrentNode = 1;
			if( mParentNode[ mCurrentNode ] == parent && this->isVisible( mCurrentNode ) ) break;
		} while( mCurrentNode != start );
	}
	this->render();
}

void 
MenuBase::select( void )
{
	if( mIgnoreNextSelection ) { mIgnoreNextSelection = false; return; }
	if( mNumberOfNodes <= 1 ) return;
	bool custom = ( mCurrentNode != 0 && this->customSelect( mCurrentNode ) );
	if( !custom )
	{
		uint16_t chosen = 0, anyChild = 0, anyYoungerSibling = 0;
		for( uint16_t candidate = mCurrentNode; candidate < mNumberOfNodes; candidate++ )
		{
			if( candidate == 0 ) continue;
			if( candidate != mCurrentNode && mParentNode[ candidate ] == mParentNode[ mCurrentNode ] ) anyYoungerSibling = candidate;
			if( mParentNode[ candidate ] != mCurrentNode ) continue;
			anyChild = candidate;
			if( !this->isVisible( candidate ) ) continue;
			if( !chosen ) chosen = candidate; // default to the first visible child
			if( this->isSelected( candidate ) ) { chosen = candidate; break; } // override and stop if you find a visible child that is also explicitly selected
		}
		custom = !anyChild && !anyYoungerSibling; // if the node has no children, visible or otherwise, and did not have a "(GO BACK)" node inserted after it (the python script does not add one if there's only one node 
	                                              // and it has a custom @left and @right), then treat it as an adjustable value---so now your menu spec does not need to explicitly include `@select: return true;`
		if( !custom && chosen ) mCurrentNode = chosen;
	}
	if( custom ) 
	{                                     
		this->render( 3 ); // flicker the selected option
		this->back();
		return;
	}
	this->render();
}

void 
MenuBase::back( void )
{
	if( mCurrentNode == 0 || mNumberOfNodes <= 1 ) return;
	if( mParentNode[ mCurrentNode ] != 0 ) mCurrentNode = mParentNode[ mCurrentNode ];
	this->render();
}


void
MenuBase::home( void )
{
	mCurrentNode = 0;
	bool previousIgnoreSetting = mIgnoreNextSelection;
	mIgnoreNextSelection = false;
	this->select();
	mIgnoreNextSelection = previousIgnoreSetting;
}

void 
MenuBase::ignoreNextSelection( bool setting )
{
	mIgnoreNextSelection = setting;
}

uint16_t 
MenuBase::getCurrentNode( void )
{
	return mCurrentNode;
}

#endif // __MenuBase_CPP__
