/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __MenuBase_H__
#define   __MenuBase_H__

#include <stdint.h>
#include <LiquidCrystal.h>

typedef  const char *    MenuLabel;
typedef  const char *    StoredMenuLabel;
//typedef  const String &  MenuLabel;
//typedef  String          StoredMenuLabel;

class MenuBase
{
	public:
		MenuBase( LiquidCrystal & lcd, uint16_t nSlots, MenuLabel title="Menu" );
		virtual ~MenuBase();
		
		virtual void init( void ); // should contain one or more AddNode() calls
		virtual bool isVisible(    uint16_t nodeIndex ); // should return true if the node in question should be visible, false if it is to be hidden
		virtual bool isSelected(   uint16_t nodeIndex ); // should return true if the node in question is the selected child of its parent
		virtual bool customRender( uint16_t nodeIndex ); // should return false if the node name should be rendered by default, but true if the method has handled rendering
		virtual bool customLeft(   uint16_t nodeIndex ); // should return false if the "left"  action should just move to the previous sibling node, but true if the method overrides this (e.g. by decrementing a variable associated with the current node)
		virtual bool customRight(  uint16_t nodeIndex ); // should return false if the "right" action should just move to the next     sibling node, but true if the method overrides this (e.g. by incrementing a variable associated with the current node)
		virtual bool customSelect( uint16_t nodeIndex ); // should return false if the "select" action should just drill down into lower menu items, true if the "select" action should actually go back up one level after peforming the custom-coded action

		void begin( void );
		void render( unsigned short nBlinks=0, unsigned long msecPerBlink=100 );
		void left( void );
		void right( void );
		void select( void );
		void back( void );
		void home( void );
		
		void ignoreNextSelection( bool setting=true );
		uint16_t getCurrentNode( void );
		
		LiquidCrystal & lcd;
		bool modifyNavigation; // if you set this to true, default (non-custom) left() behavior will be to go back(), and default (non-custom) right() behavior will be to select. One good use for it is to set it to reflect whether the selector button is pressed: press-and-twist-left to go back.
	
	protected:
		uint16_t addNode( uint16_t parentNode, MenuLabel text );
		
	private:
		uint16_t           mMaxNumberOfNodes;
		uint16_t           mNumberOfNodes;
		uint16_t           mCurrentNode;
		bool               mIgnoreNextSelection;
		uint16_t *         mParentNode;
		StoredMenuLabel *  mString;
};
#endif // __MenuBase_H__
