/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __PersistentSettings_H__
#define   __PersistentSettings_H__

/*
Example usage::

	//#define SETTINGS_SDCARD BUILTIN_SDCARD  // leave commented-out to use EEPROM; uncomment to use the SD card (this definition must happen before #include "PersistentSettings.h")
	#include "PersistentSettings.h"
	
	// Now write your own function:
	bool PersistentSettings( bool writing )      // Your function must return a bool.
	{
		OPEN_SETTINGS( "mysketch-v1", writing ); // hard-code a filename (or EEPROM signature) here that is unique to this sketch and to these particular settings with these particular types in this particular order
		SETTING( gSomeGlobalObject.someMemberVariable ); // pass references to whatever global objects you have. This expression will always evaluate to `true` in EEPROM mode,
		// etc                                           // but in SDCARD mode it will evaluate to `false` if the read/write operation failed.
		return true; // indicates success
	}
	// Now call PersistentSettings(false) to load settings, or PersistentSettings(true) to save them.
	// Alternatively use the definitions PersistentSettings(LOAD_SETTINGS) and PersistentSettings(SAVE_SETTINGS).
	// The OPEN_SETTINGS macro will cause PersistentSettings(LOAD_SETTINGS) to return false if it failed: a silent
	// failure (no accompanying Serial message) indicates that there is no pre-existing file of the specified
	// name (or, equivalently, that the EEPROM begins with a signature that differs from the specified "filename").
	
*/


#ifdef SETTINGS_SDCARD

#	include <SD.h>

#	define OPEN_SETTINGS( FILENAME, WRITING ) \
		bool _storage_write_mode = WRITING; \
		if( !SD.begin( SETTINGS_SDCARD ) ) { Serial.println( "# SD.begin() failed" ); return false; } \
		File _storage = SD.open( FILENAME, _storage_write_mode ? FILE_WRITE : FILE_READ ); \
		if( !_storage && _storage_write_mode ) Serial.println( "# SD.open() failed to open settings file for writing" ); \
		if( !_storage ) return false;
		
#	define SETTING( X ) \
		( sizeof( X ) == ( _storage_write_mode ? _storage.write(  ( void * )&X, sizeof( X ) ) : _storage.read(  ( void * )&X, sizeof( X ) ) ) )

#else // SETTINGS_SDCARD is undefined, so let's use EEPROM (but be aware of its lifetime limit of ~100,000 write cycles - don't call these routines frequently!)

#	include <EEPROM.h>

	int _init_EEPROM( const String & signature, bool writing )
	{
		const char * sig = signature.c_str();
		for( size_t i = 0; i <= signature.length(); i++ ) // include the null terminator
		{
			if( EEPROM.read( i ) != sig[ i ] )
			{
				if( writing ) EEPROM.write( i, sig[ i ] );
				else return -1; // we have opened the "file" for reading, but it's the wrong file (has the wrong signature)
			}
		}
		return signature.length() + 1;
	}


#	define OPEN_SETTINGS( FILENAME, WRITING )  \
		bool _storage_write_mode = WRITING; \
		int _storage = _init_EEPROM( FILENAME, _storage_write_mode ); \
		if( _storage < 0 ) return false;
		
#	define SETTING( X ) \
		( _storage += ( ( _storage_write_mode ? EEPROM.put( _storage, X ) : EEPROM.get( _storage, X ) ), sizeof( X ) ),  true )

#endif // SETTINGS_SDCARD

const bool LOAD_SETTINGS = false;
const bool SAVE_SETTINGS = true;


#endif // __PersistentSettings_H__
