/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __Triggering_H__
#define   __Triggering_H__

#include <Arduino.h>
#include <RecentMax.h>

class Trigger
{
	public:
		Trigger( int inputPin, int outputPin1, int outputPin2=-1, double memoryInSeconds=0.0 );		
		virtual ~Trigger();
		
		// Sets pin modes. Usually does not need to be called explicitly since it is called during construction.
		void begin( void );
		
		// Reserves a new `RecentMax<int>` member object (or resizes the memory of any existing one) to compute the maximum signal over a sliding window. Called automatically during construction if you gave a non-zero value for the `memoryInSeconds` constructor argument.
		void rememberSamples( double maxMemoryInSeconds=1.0, unsigned long nBins=20 );

		// Reads a single signal value for comparison against `threshold`. Overshadow this virtual method in your subclasses. The base class is an AC trigger that measures the absolute value of the difference between consecutive samples.
		virtual int getSignal( void );
		
		// `t` should be the timestamp of the current `loop`, obtained from `micros()`. This method calls `read()` and then (if appropriate) `write()`, all in one call. It returns the resulting output-pulse state.
		inline bool process( unsigned long t ) { return read( t ) ? write( t ) : mPulseOn; }
		
		// `t` should be the timestamp of the current `loop`, obtained from `micros()`. If it's time to take a new reading, this method takes one using the `getSignal()` method and returns `true`. Otherwise it returns `false`. Note that `process()` calls `read()` and `write()` in one go.
		bool read( unsigned long t );
		
		// `t` should be the timestamp of the current `loop`, obtained from `micros()`. If the last call to `read()` took a new reading, this method processes it, changing the output-pulse state if appropriate. It returns the resulting output-pulse state. Note that `process()` calls `read()` and `write()` in one go.
		bool write( unsigned long t );
		
		// `t` should be the timestamp of the current `loop`, obtained from `micros()`. Use this to synchronize sampling to the current timestamp, and to set the state of the input signal in software. You would use `force(t, state) && write(t)` instead of `read(t) && write(t)` to drive a trigger by some software event while still allowing that event to be processed by the `Trigger` instance's pulse-sustaining/debouncing logic.
		bool force( unsigned long t, bool exceedThreshold );

		// The truth value of a `Trigger` instance reflects whether or not the output pulse is currently on.
		operator bool() const { return mPulseOn; }
		
		unsigned long samplePeriodInMicros; // sample period in microseconds
		bool inverted; // If set to `true`, the `Trigger` triggers when the signal goes *below* the `threshold` level rather than above (not really useful in the `Trigger` base class, but can be useful in `TriggerDC` and potentially other subclasses).
		int threshold; // When the `getSignal()` output exceeds this value, an output pulse is triggered.
		int recentMax; // This diagnostic value can help in setting the threshold. Note that this it only be non-zero if you have specified `memoryInSeconds` > 0 in the constructor, or have otherwise used `rememberSamples()` to configure memory for recent signal values.
		unsigned long minimumPulseDurationInMicros; // The output pulse will remain on for at least this many microseconds after the first threshold-crossing.
		unsigned long pulseDurationIncrementInMicros; // The output pulse will continue until at least this many microseconds after the most recent threshold-crossing.
		unsigned long refractoryPeriodInMillis; // This is the refractory period in milliseconds, for de-bouncing output pulse onsets.
		void ( *pulseStartCallback  )( Trigger& t ); // This optional pointer to a `void callback(Trigger &)` function is called whenever a trigger pulse starts.
		void ( *pulseFinishCallback )( Trigger& t ); // This optional pointer to a `void callback(Trigger &)` function is called whenever a trigger pulse finishes.
		
		int mInputPin;
		int mOutputPin1;
		int mOutputPin2;
		int mPreviousReading;
		int mSignal;
		bool mPulseOn;
		bool mInRefractoryPeriod;
		unsigned long mLastSampleTimeInMicros;
		unsigned long mTimeLastExceeded;
		unsigned long mLastPulseStartTimeInMicros;
		unsigned long mLastPulseFinishTimeInMicros;
		unsigned long mLastPulseDurationInMicros;
		unsigned long mPulsesStarted;
		unsigned long mPulsesFinished;
		RecentMax< int > * mSlidingWindow;
		
};

class TriggerDC : public Trigger
{
	public:
		TriggerDC( int inputPin, int outputPin1, int outputPin2=-1, double memoryInSeconds=0.0 );
		virtual ~TriggerDC() override;
		
		// For a `TriggerDC` this is a straightforward `analogRead()`.
		virtual int getSignal( void ) override;
};

#endif // __Triggering_H__
