#!/usr/bin/env -S python  #

# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# Development was supported by the NIH, NYS SCIRB, Veterans Affairs RRD,
# and the Stratton VA Medical Center.
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$

"""
Translates a text representation of a menu system into a C++ (Arduino)
header that subclasses `MenuBase` as `Menu`. 

If no input files are specified, a demo menu, compatible with the example
sketch `MenusAndSettings.ino`, is produced.
"""

DEMO_SPEC = r"""

# This menu specification gets used if you call the script without specifying any
# input files. It is compatible with the example sketch `MenusAndSettings.ino`,
# which defines the global variables referenced below as well as the global
# functions `Adjust()` and `Settings()`. The other references it uses, `lcd` and
# `modifyNavigation`, are member variables of the `MenuBase` superclass.

/Speaker Config/STEREO               @select:  gConfig = 2.0;   @selected: (gConfig == 2.0)
/Speaker Config/5.1 SURROUND         @select:  gConfig = 5.1;   @selected: (gConfig == 5.1) 
/Speaker Config/7.1 SURROUND         @select:  gConfig = 7.1;   @selected: (gConfig == 7.1)
\Volume\L/R Front\x                  @visible:                    @render: lcd.print(gFront       );   @left: return Adjust(-1, gFront,        modifyNavigation);   @right: return Adjust(+1, gFront,        modifyNavigation);
\Volume\L/R Surround\x               @visible: (gConfig >= 5.1)   @render: lcd.print(gSurround    );   @left: return Adjust(-1, gSurround,     modifyNavigation);   @right: return Adjust(+1, gSurround,     modifyNavigation);
\Volume\L/R Back Surr'nd\x           @visible: (gConfig >= 7.1)   @render: lcd.print(gBackSurround);   @left: return Adjust(-1, gBackSurround, modifyNavigation);   @right: return Adjust(+1, gBackSurround, modifyNavigation);
\Volume\Center\x                     @visible: (gConfig >= 5.1)   @render: lcd.print(gCenter      );   @left: return Adjust(-1, gCenter,       modifyNavigation);   @right: return Adjust(+1, gCenter,       modifyNavigation);
\Volume\Subwoofer\x                  @visible: (gConfig >= 5.1)   @render: lcd.print(gSubwoofer   );   @left: return Adjust(-1, gSubwoofer,    modifyNavigation);   @right: return Adjust(+1, gSubwoofer,    modifyNavigation);
/Stored Settings/SAVE NOW            @select: Settings(SAVE_SETTINGS);
/Stored Settings/REVERT TO SAVED     @select: Settings(LOAD_SETTINGS);
"""  

import os
import re
import sys
import textwrap

class Node:
	def __init__( self, parent, string, code=None ):
		self.parent, self.string = parent, string
		self.code = code if code else {}
	def __repr__( self ):
		return '%s(%r, %r%s)' % ( self.__class__.__name__, self.parent, self.string, ( ', code=...' if self.code else '' ) )
	
def Indent( nLevels, text ):
	return ( '\n' + text ).replace( '\n', '\n' + '\t' * nLevels ).rstrip().lstrip( '\n' )

def FormatCPP( text, **kwargs ):
	text = text.lstrip( '\n' )
	indentChars = text[ : len( text ) - len( text.lstrip() ) ]
	text = ( '\n' + text ).replace( '\n' + indentChars, '\n' ).strip()
	for find, replace in kwargs.items(): text = text.replace( '@' + find + '@', str( replace ) )
	return text

def FormatSnippet( text, expectReturnExpression=False, addReturn='true', singleLineLimit=40 ):
	text = FormatCPP( text )
	if expectReturnExpression:
		if not re.findall( r'^return[\s\+\-\(\)\~\'\"\.]', text ): text = 'return ' + text;
		if not text.endswith( ';' ): text += ';'
	elif addReturn and not re.findall( r'return[\s\+\-\(\)\~\'\"\.][^;]+;$', text ):
		text += ' return %s;' % addReturn;
	if len( text ) > singleLineLimit or '\n' in text:
		text = SplitCPP( text, rejoin='\n', isolateBraces=True )
	return text;

def Wrap( text, nCols=80 ):
	return '\n'.join( textwrap.wrap( text, nCols ) )

class Menu:
	
	SHORTCUTS = dict(
		isVisible    = 'visible',
		isSelected   = 'selected',
		customRender = 'render',
		customLeft   = 'left',
		customRight  = 'right',
		customSelect = 'select',
	)
	
	def __init__( self, spec='', title='Menu', back='(GO BACK)' ):
		self.title = title
		self.back = back
		self.byIndex = {}
		self.byPath = {}
		self.Add( '/' )
		self.byIndex[ 0 ].parent = None
		self.Add( spec )
		
	def Add( self, line ):
		line = line.strip()
		if '\n' in line: return [ self.Add( line ) for line in spec.split( '\n' ) if line.strip() ]
		if not line: return 0
		if line.startswith( '#' ): return 0
		if line.startswith( '@' ): return 0
		pieces = re.split( r'\s+@(' + '|'.join( self.SHORTCUTS.values() ) + '):', line )
		code = { k : v.strip() for k, v in zip( pieces[ 1::2 ], pieces[ 2::2 ] ) if v.strip() }
		path = pieces[ 0 ].strip()
		delim = path[ 0 ]
		path = tuple( path.rstrip( delim ).split( delim ) )
		existingNumber = self.byPath.get( path, None )
		if existingNumber is not None:
			self.byIndex[ existingNumber ].code.update( code )
			return existingNumber
		parent = 0
		for i in range( 1, len( path ) ):
			parentPath = path[ : i ]
			name = path[ i ]
			parent = self.byPath.get( parentPath, None )
			if parent is None: parent = self.Add( delim.join( parentPath ) )			
		nodeIndex = len( self.byIndex )
		node = Node( parent, path[ -1 ], code )
		self.byIndex[ nodeIndex ] = node
		self.byPath[ path ] = nodeIndex
		return nodeIndex
				
	def AllItems( self ):
		items = sorted( self.byIndex.items() )
		parents = {}
		# compile a list of parents (exclude the root, node 0, because that will need neither custom isVisible() code nor a 'GO BACK' child)
		for nodeIndex, node in items:
			if node.parent: parents.setdefault( node.parent, [] ).append( node )
		# propagate custom isVisible() code from only-children to their parents (if the parent has no such customization). Do it recursively if the parent is also an only-child.
		for nodeIndex, node in items[ ::-1 ]:
			visCode = node.code.get( self.SHORTCUTS[ 'isVisible' ], None )
			if not visCode: continue
			if node.parent == 0: continue
			nChildrenOfParent = len( parents[ node.parent ] )
			if nChildrenOfParent != 1: continue
			parentNode = self.byIndex[ node.parent ]
			parentVisCode = parentNode.code.get( self.SHORTCUTS[ 'isVisible' ], None )
			if parentVisCode: continue
			parentNode.code[ self.SHORTCUTS[ 'isVisible' ] ] = visCode
		# parents of with only one child, where the child overrides customLeft and customRight, do not need a 'GO BACK' child
		for parentIndex, children in list( parents.items() ):
			if len( children ) == 1 and children[ 0 ].code.get( 'left', None ) and children[ 0 ].code.get( 'right', None ):
				parents.pop( parentIndex )
		# add 'GO BACK' children to everyone else
		items += enumerate( [ Node( parent, self.back, code={ self.SHORTCUTS[ 'customSelect' ]: 'return true;' } ) for parent in sorted( parents ) ], len( items ) )
		return items
		
	def CollateCode( self, key, expectReturnExpression=False ):
		collated = {}
		for nodeIndex, node in self.AllItems():
			code = node.code.get( key, '' )
			if not code: continue
			collated.setdefault( code, [] ).append( 'case %d:' % nodeIndex )
		return '\n'.join( Wrap( ' '.join( cases ) ) + '\n' + Indent( 1, FormatSnippet( code, expectReturnExpression=expectReturnExpression, addReturn='true' ) ) for code, cases in collated.items() )
	
	def Depth( self, node ):
		depth = 0
		while node.parent is not None: node = self.byIndex[ node.parent ]; depth += 1
		return depth
		
	def InjectCases( self, template, key, expectReturnExpression=False, nIndents=2 ):
		code = self.CollateCode( key, expectReturnExpression=expectReturnExpression )
		if not code: return ''
		return FormatCPP( template, cases=Indent( 2, code ).strip() )
	
	@property
	def code( self ):
		return FormatCPP( """
			/*
			 * This file was auto-generated by `@utility@`.
			 * It is designed to be `#include`d in between your global declarations and your `setup()`, because
			 * any custom code in this file will presumably need to refer to those sketch-specific globals.
			*/
			 
			#include "MenuBase.h"
			class Menu: public MenuBase
			{
				public:
					Menu( LiquidCrystal & lcd ) : MenuBase( lcd, @size@, "@title@" ) { init(); }
					virtual ~Menu() {}
					@methods@
			};
		""", utility=os.path.basename( __file__ ), size=len( self.AllItems() ), title=self.title, methods=Indent( 2, self.methods ).strip() ) + '\n'
	@property
	def methods( self ):
		return '\n'.join( method for methodName in [ x for x in dir( self ) if x.startswith( 'method_' ) ] for method in [ getattr( self, methodName ) ] if method.strip() )
		
	def NodeStringLiteral( self, node, indent='  ' ):
		if not indent: indent = ''
		return indent * ( self.Depth( node ) - 1 ) + '"' + node.string.replace( '\\', '\\\\' ).replace( '"', '\\"' ) + '"'
		
	@property
	def method_init( self ):
		statements = '\n'.join(
			#'static const char node{nodeIndex:03d}[] PROGMEM = {label:30s} addNode({parentIndex: 3d}, node{nodeIndex:03d});'.format( nodeIndex=nodeIndex, parentIndex=node.parent, label=self.NodeStringLiteral( node ) + ';' )
			'/* node {nodeIndex: 4d} */ addNode( {parentIndex: 4d}, {label:30s} );'.format( nodeIndex=nodeIndex, parentIndex=node.parent, label=self.NodeStringLiteral( node ) )
			for nodeIndex, node in self.AllItems()
			if node.parent is not None
		)
		return FormatCPP( """
			virtual void init( void ) override
			{
				@nodes@
			}
		""", nodes=Indent( 1, statements ).strip() )
	@property
	def method_isVisible( self ):
		return self.InjectCases( """
			virtual bool isVisible( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return true;
				}
			}
		""", self.SHORTCUTS[ 'isVisible' ], expectReturnExpression=True )
	@property
	def method_isSelected( self ):
		return self.InjectCases( """
			virtual bool isSelected( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return false;
				}
			}
		""", self.SHORTCUTS[ 'isSelected' ], expectReturnExpression=True )
	@property
	def method_customSelect( self ):
		return self.InjectCases( """
			virtual bool customSelect( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return false;
				}
			}
		""", self.SHORTCUTS[ 'customSelect' ] )
	@property
	def method_customLeft( self ):
		return self.InjectCases( """
			virtual bool customLeft( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return false;
				}
			}
		""", self.SHORTCUTS[ 'customLeft' ] )
	@property
	def method_customRight( self ):
		return self.InjectCases( """
			virtual bool customRight( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return false;
				}
			}
		""", self.SHORTCUTS[ 'customRight' ] )
	@property
	def method_customRender( self ):
		return self.InjectCases( """
			virtual bool customRender( uint16_t nodeIndex ) override
			{
				switch( nodeIndex )
				{
					@cases@
					default:
						return false;
				}
			}
		""", self.SHORTCUTS[ 'customRender' ] )


def SplitCPP( s, isolateBraces=True, rejoin=None ):
	inQuote = None
	brackets = ''
	backslash = False
	comment = ''
	parts = [ '' ]
	for char in s:
		parts[ -1 ] += char
		if parts[ -1 ].isspace(): parts[ -1 ] = ''
		if inQuote and not backslash and char == '\\': backslash = True; continue
		if inQuote and backslash: backslash = False; continue
		if char == inQuote: inQuote = None; continue
		if char in [ '"', "'" ] and comment not in ( '//', '/*', '/**' ): inQuote = char; continue
		if inQuote: continue
		if not comment and char == '/': comment = '/'
		elif comment == '//':
			if char == '\n': comment = ''
			else: continue
		elif comment == '/' and char in '/*': comment += char
		elif comment == '/*' and char == '*': comment += '*'
		elif comment == '/**' and char == '/': comment = ''
		elif comment == '/**': comment = '/*'
		if comment in [ '/*', '/**' ] and char != '\n': continue
		if char in '({[': brackets += char;
		if brackets.endswith( '(' ) and char == ')': brackets = brackets[ :-1 ];
		if brackets.endswith( '[' ) and char == ']': brackets = brackets[ :-1 ];
		if brackets.endswith( '{' ) and char == '}': brackets = brackets[ :-1 ];
		if char == '\n' or ( char == ';' and '(' not in brackets ): parts[ -1 ] = parts[ -1 ].rstrip(); parts.append( '' )
		if char in '{}' and isolateBraces:
			parts[ -1 ] = parts[ -1 ][ :-1 ].rstrip();
			while parts and not parts[ -1 ]: parts.pop( -1 )
			parts.append( char )
			parts.append( '' )
	if not parts[ -1 ]: parts.pop( -1 )
	if rejoin is None: return parts
	indents = 0;
	for i, part in enumerate( parts ):
		part = part.strip()
		if part.startswith( tuple( '])}' ) ): indents -= 1
		tabs = '\t' * indents
		if part.endswith( tuple( '{([' ) ): indents += 1
		parts[ i ] = tabs + part
		# TODO: - maybe put '{ one; two; }' back together again if it is short enough?
		#       - 'one; /* comment */ two;' gets split before the comment but not after
	return rejoin.join( parts )
	
if __name__ == '__main__':
	spec = DEMO_SPEC
	import argparse
	parser = argparse.ArgumentParser( description=__doc__, formatter_class=argparse.RawTextHelpFormatter )
	parser.add_argument(       "filenames", nargs='*', metavar='INPUT_FILENAME', help='menu specification file(s)---use `-` for stdin' )
	parser.add_argument( "-o", "--output",  metavar='OUTPUT_FILENAME',  default='-', help='output file (defaults to `-`, which means use stdout)' )
	OPTS = parser.parse_args()
	def ReadContents( filename ):
		if filename == '-': return sys.stdin.read()
		with open( filename ) as fh: return fh.read()
	if OPTS.filenames: spec = '\n'.join( ReadContents( filename ) for filename in OPTS.filenames )
	m = Menu( spec )
	if OPTS.output == '-':
		print( m.code )
		sys.stderr.write( 'Header content written to stdout.\n' )
	else:
		OPTS.output = os.path.abspath( os.path.expanduser( OPTS.output ) )#.replace( os.path.sep, '/' )
		with open( OPTS.output, 'wt' ) as fh: fh.write( m.code )
		sys.stderr.write( 'Header content written to %s\n' % OPTS.output )
	sys.stderr.write( '%d bytes used for string labels\n' % sum( len( node.string ) for nodeIndex, node in m.AllItems() if node.parent is not None ) )
